/**
 * Surveyengine Map
 *
 * extends leaflet L.Map class
 *
 * addtional methods:
 *
 *  - setUserMarker
 *  - setLeisureMarker
 *
 *  @version 2.5
 *
 *  @author abernhard <a.bernhard@23go.de>
 *
 *** Changelog
 *  2.5     2020-04-15
 *   - added "token" as resource element
 *   - fixed "reload" icon
 *
 *  2.4     2018-10-08
 *   - added try-catch to gracefully handle initialization errors
 *
 *  2.3     2018-07-23
 *   - set fitBounds options {animate:false} as it kept colliding with the zoom-to-zip methods
 *
 *  2.2     2018-07-19
 *   - added ".stop()" before flytobounds on showPlz to prevent animation collision with loaded-center animations
 *
 *  2.1     2018-07-11
 *   - catch try to resolve some leaflet errors on .remove
 *
 *  2.0     2018-07-02
 *   - added "reload map" control button
 *   - added "ready" parameter for map initialition methods
 *
 *  1.4     2018-05-30
 *   - unfixed initially not missing Boundingbox handling
 *   - fixed race-condition on init/showplz/fitbounds
 *   - fit circle/plz feature flyToBounds
 *   - fixed no-shape, no-circle getBounds
 *
 *  1.3     2018-05-29
 *   - updated default value handling for SE_PLZ_DATA_URI, SE_MAP_STYLE
 *   - fixed missing Boundingbox handling
 *   - added "debug" parameter
 *   - re-added default min/maxZoom
 *
 *  1       2018-05-18
 *   - First version
 *
 */

/**
 * include in script
 *
 *
 <link rel="stylesheet" href="leaflet.1.2.0.css">
 <script src="leaflet.1.2.0.js"></script>

 <!-- Mapbox GL -->
 <link rel="stylesheet" href="mapbox_gl.css">
 <script src="mapbox_gl.0.14.3.js"></script>
 <script src="leaflet_mapbox_gl.js"></script>
 <script src="se_map.v1.js"></script>
 <link rel="stylesheet" href="se_map.v1.css">

 <div id="se_map1" class="se_map smp_loading" data-loading-text="Karte wird geladen ..."></div>
 <div style="display: none;" id="seMap_plz"><span class="expression">$plz</span></div>

 <script>
 ! function()
 {

   var userPlzVal = $('#seMap_plz').text(),
     inputUserPosSelector = '*[name=r##latlng_home]',
     mapCenterCoordinates = [52.817146, 9.527094];

   // ---------- init -------------------

   var se_map1 = L.SeMap('se_map1',
   {
     //riverHighlights: 'West', // West, Ost, Zwoenitz .. oder false               // optional
     // area: "aller", // aller, donau, nahe, sachsen,                             // optional
     resources: {
        styleFile: "some/other/file.than.default" or mapbox style link
        token: [optional mapbox style token]                                       // optional
     },
     //center: mapCenterCoordinates // optional
     whenReady: function(map){

       // add or remove this part to set a user position on the map
       map.on('click', function(ev)
       {
         var _map = this,
           latlng = ev.latlng;

         _map.setUserMarker(latlng);

         var latlngStr = latlng.lat + "," + latlng.lng;

         $(inputUserPosSelector).val(latlngStr)

       });

       // restore previously set data
       setTimeout(function()
       { // necessary timeout due to async element rendering
         if ($(inputUserPosSelector).val())
         {
           map.setUserMarker($(inputUserPosSelector).val().split(","));
         }
       }, 250);



       map.showPlz(userPlzVal,
       {
         error: function()
         {
           se_map1.setView(mapCenterCoordinates);
           alert('Die angegebene PLZ (' + userPlzVal +
             ') konnte nicht gefunden werden.');
         }
       });

     }
   });



 }();
 </script>
 *
 *
 */

(function(window){

    var LOG_LABEL = "SE_MAP";
    var L = window.L;

    if(!L){
        console.error(LOG_LABEL, "Leaflet not found", {L: L});
        return;
    }

    
    var utils = {};
    
    utils.merge = function(a, b){
        var obj = {};
        
        for(var i in arguments){
            for(var i2 in arguments[i]){
                obj[i2] = arguments[i][i2];
            }
        }        
        return obj;
    };


    // ---------- SE Map -----------------

    var SE_PLZ_DATA_URI_DEFAULT = "data/",
        SE_MAP_STYLE_DEFAULT = "se_map_style.default.json",
        SE_PLZ_DATA_URI, SE_MAP_STYLE;


    var SE_MAP_EVENTS = {
            LOADED: "seMap_loaded"
        };



    L.SeMap = L.Map.extend({
        initialize: function(id, options){
            var _map = this;

            SE_PLZ_DATA_URI = window.SE_PLZ_DATA_URI || SE_PLZ_DATA_URI_DEFAULT;
            SE_MAP_STYLE = window.SE_MAP_STYLE || SE_MAP_STYLE_DEFAULT;

            // -------- read in defaults

            options = utils.merge({
                plzDataUri: SE_PLZ_DATA_URI,
                zoom: 10,
                attributionControl: false,
                minZoom: 8,
                maxZoom: 15
            }, options || {});

            if(options.debug){
                options.minZoom = null;
                options.maxZoom = null;
                console.info(LOG_LABEL, "DEBUG=TRUE", "disabling min max zoom restrictions");
            }


            if(options.area && window.SE_MAP_AREAS){
                options.resources = utils.merge(options.resources || {}, SE_MAP_AREAS[options.area] || {});
            }


            options.resources = utils.merge({
                styleFile: SE_MAP_STYLE,
                //boundingBox: [[48.49591, 10.61543] , [48.20702, 11.17464]],
                plzDataUri: SE_PLZ_DATA_URI
            }, options.resources || {});


            // --------- init map

            options.id = id;
            _map.node = L.DomUtil.get(id);

            _map.node.classList.add('smp_loading');


            if(options.maxBounds){
                options.maxBounds = L.latLngBounds(options.maxBounds);
            }else if(options.resources.boundingBox){
                options.maxBounds = L.latLngBounds(options.resources.boundingBox).pad(.75);
            }


            if(options.maxBounds){
                if(!options.center || !options.maxBounds.contains(options.center)){
                    console.info(LOG_LABEL, "set default center to bounding box center");
                    options.center = options.maxBounds.getCenter();
                }

            }else if(!options.center){
                options.center = [0,0];
            }

            if(options.debug){
                options.maxBounds = null;
                console.info(LOG_LABEL, "DEBUG=TRUE", "disabling max bounds restrictions");
            }

            _map._options = options;

            try{

                L.Map.prototype.initialize.call(_map, id, options);

            }catch(err){
                if(err.toString().match(/webgl/)){
                    alert('Failed to load WebGL. Please restart your browser and try again!');

                }else if(err.toString().match(/Map container is already initialized/)){

                    console.warn(LOG_LABEL, "reloading map", err);
                    if(!_map.isReloading){
                        _map._reload();
                    }else{
                        console.error(LOG_LABEL, "reloading failed", err);
                    }

                }else{
                    console.error(LOG_LABEL, "error on init", err);
                }
            }


            L.mapboxGL({
                style: options.resources.styleFile,
                accessToken: options.resources.token || 'no-token'
            }).addTo(_map);

            L.control.scale({
                metric: true,
                imperial: false
            }).addTo(_map);

            _map._createReloadButton();


            if(options.resources.layers){
                _map._initLayers().then(function(){
                    _map._setMapLoaded();
                });
            }else{
                _map._setMapLoaded();
            }

        },
        initCenter: null, //

        _reload: function(){
            var _map = this,

                mapId = _map._options.id + "",
                options = utils.merge({}, _map._options);

            setTimeout(function(){
                _map = new L.SeMap(mapId, options);
            }, 250);

            _map.stop();
            _map.isReloading = true;
            try{ _map.remove(); }catch(err){ /* ignore */ }
            _map.isLoaded = false;

        },

        _createReloadButton: function(){

            var _map = this;

            L.Control.Reload = L.Control.extend(
                {
                    options:
                        {
                            position: 'topright'
                        },
                    onAdd: function (map) {
                        var controlDiv = L.DomUtil.create('div', 'leaflet-draw-toolbar leaflet-bar');
                        L.DomEvent
                            .addListener(controlDiv, 'click', L.DomEvent.stopPropagation)
                            .addListener(controlDiv, 'click', L.DomEvent.preventDefault)
                            .addListener(controlDiv, 'click', function () {

                                _map._reload();

                            });

                        var controlUI = L.DomUtil.create('a', 'leaflet-control-zoom-in', controlDiv);
                        controlUI.title = 'Reload map';
                        controlUI.innerHTML = "&#8635;";
                        controlUI.href = '#';

                        return controlDiv;
                    }
                });

            var reloadControl = new L.Control.Reload();
            _map.addControl(reloadControl);

        },

        _setMapLoaded: function(){
            var _map = this;

            if(_map.isLoaded) return;

            var maxIntervals = 100,
                waitForMapVisible = setInterval(function(){
                    if(!maxIntervals){
                        clearInterval(waitForMapVisible);
                        alert("Karte konnte nicht initialisiert werden.");
                        return;
                    }

                    if($(_map.node).is(':visible')){
                        clearInterval(waitForMapVisible);

                        setTimeout(function(){
                            // remove the .smp_loading class off the map dom node
                            L.DomUtil.removeClass(_map.node, "smp_loading");

                            // force a redraw after everything is loaded
                            _map.invalidateSize();

                            if(_map.initCenter){
                                _map.initCenter();
                            }else if(_map._options.maxBounds){
                                _map.fitBounds(_map._options.maxBounds, {animate: false});
                            }


                            if("function" === typeof _map._options.whenReady){

                                _map._options.whenReady.apply(_map, [_map]);

                            }

                            _map.isLoaded = true;
                            _map.isReloading = false;
                            _map.fire(SE_MAP_EVENTS.LOADED);

                        }, 1250);

                        return;
                    }

                    maxIntervals--;
                }, 200);



        },

        showPlz: function(plzStr, options){
            var _map = this;

            options = utils.merge({
                hideShape: false,
                showCircle: false, // {radius: INT_IN_METER, color}
                success: function(){},
                error: function(){}
            }, options || {});

            if(!_map._options.resources.plzDataUri){
                console.error(LOG_LABEL, "PLZ data uri missing");
                return;
            }

            $.ajax({
                dataType: "json",
                url: _map._options.resources.plzDataUri + plzStr,
                success: function (data) {

                    var featureStyleDefault = {color: "#ffd600", weight: 2, pane: "shadowPane"},
                        plzFeatureStyle = utils.merge({}, featureStyleDefault),
                        plzFeature,
                        plzFeatureLayer = L.featureGroup();
                        plzFeatureLayer.addTo(_map);

                    plzFeature = L.geoJSON(data, plzFeatureStyle);
                    if(!options.hideShape){
                        plzFeature.addTo(plzFeatureLayer);
                    }

                    if(parseInt(options.showCircle)){
                        var circleRadius = parseInt(options.showCircle),
                            circleCenter = plzFeature.getBounds().getCenter(),
                            circleStyle = utils.merge(featureStyleDefault, {radius: circleRadius}),
                            circleFeature = L.circle(circleCenter, circleStyle);

                        circleFeature.addTo(plzFeatureLayer);

                    }

                    var featureBounds = plzFeatureLayer.getBounds();
                    if(!featureBounds.isValid()){
                        featureBounds = plzFeature.getBounds();
                    }

                    _map.initCenter = function(){
                        setTimeout(function(){
                            _map.stop();
                            _map.flyToBounds(featureBounds);
                        }, 100);
                    };

                    if(_map.isLoaded){
                        _map.initCenter();
                    }

                    options.success(plzStr);
                },
                error: function(err){

                    console.warn(LOG_LABEL, "PLZ not found", plzStr, err, arguments);
                    options.error(plzStr);

                }
            });
        },

        _setMarker: function (marker, latLng) {
            var _map = this;
            marker.setLatLng(latLng);

            if(!marker.isOnMap){
                marker.isOnMap = true;

                marker.addTo(_map);
            }

            return marker;

        },
        _userMarker: L.marker([0,0], {icon: L.icon({
            iconUrl: 'marker2.png',
            iconSize: [31, 30],
            iconAnchor: [12, 30]
        }), draggable: false, pane: "markerPane"}),
        setUserMarker: function(latLng){
            var _map = this;

            var relPosLatLng = latLng; //L.GeoJSON.coordsToLatLng(L.GeoJSON.latLngToCoords(latLng, 2.4));


            return _map._setMarker(_map._userMarker, relPosLatLng);
        },
        _leisureMarker: L.marker([0,0], {icon: L.icon({
            iconUrl: 'marker.png',
            iconSize: [31, 30],
            iconAnchor: [12, 30]
        }), draggable: false, pane: "markerPane"}),
        setLeisureMarker: function(latLng){
            var _map = this;
            return _map._setMarker(_map._leisureMarker, latLng);
        },

        getEpsg3857Bounds: function(){
            var _map = this,

                centerLatLng = _map.getCenter(),
                centerPnt = L.CRS.EPSG3857.project(centerLatLng),

                boundsLatLng = _map.getBounds().pad(0.1),
                boundsPnt = [
                    L.CRS.EPSG3857.project(boundsLatLng.getSouthWest()),
                    L.CRS.EPSG3857.project(boundsLatLng.getNorthEast())
                ],

                zoomVal = _map.getZoom(),

                epsg3857bounds = [boundsPnt[0].y, boundsPnt[0].x, boundsPnt[1].y, boundsPnt[1].x];

            return epsg3857bounds;
        },

        _geoLayers: {
            displayed: L.featureGroup({interactive: true}),
            all: [],
            layers: {}
        },
        _initLayers: function(){
            var _map = this,
                _res = $.Deferred(),
                layersCount = 0,
                layersAdded = 0;


            var layers;
            try{ layers = _map.options.resources.layers; }catch(err){}
            if(!layers) return;

            _map._geoLayers.displayed.addTo(_map);

            for(var geoLayerId in layers){
                layersCount++;
                var geoLayerSettings = layers[geoLayerId];

                _map.addGeoLayer(geoLayerId, geoLayerSettings.uri, geoLayerSettings.options).then(function(){
                    layersAdded++;

                    if(layersCount === layersAdded){
                        _res.resolve();
                    }
                });
            }

            _map.on("moveend zoomend resize viewreset", _map._updateGeoLayers);

            return _res.promise();

        },
        addGeoLayer: function(id, uri, options){
            var _map = this,
                _res = $.Deferred();

            options = options || {};

            options.style = utils.merge({
                color: "yellow",
                weight: 2,
                smoothFactor: 1.5
            }, options.style || {});

            var onEachFeatureCb = options.onEachFeature || function(){};
            options.onEachFeature = function(feature, layer){

                // prepped for later.

                return onEachFeatureCb(feature, layer);
            };

            $.ajax(uri, {
                method: "get",
                dataType: "json",
                error: function(err){
                    console.error(LOG_LABEL, "error reading geojson file " + geoJsonFileUri, err, arguments);
                    _res.reject();
                },
                success: function(data) {

                    var geoLayer = L.geoJson(data, options);
                    _map._geoLayers.layers[id] = geoLayer;


                    geoLayer.eachLayer(function(layer){
                        layer.bounds = layer.bounds || layer.getBounds();
                        _map._geoLayers.all.push(layer);
                    });

                    geoLayer.bounds = geoLayer.getBounds();

                    _map._updateGeoLayers();
                    _res.resolve();

                }
            });

            return _res.promise();

        },

        _updateGeoLayers: function(){
            var _map = this;

            // get current bounding box
            var currentBounds = _map.getBounds();

            // loop through all geo layers
            var _tiles = _map._geoLayers.all,
                _tiles_displayed = _map._geoLayers.displayed,
                tile;

            for(var i in _tiles){
                tile = _tiles[i];

                tile.bounds = tile.bounds || tile.getBounds();

                // if tile is in view -> add to its .cityLayer
                if(tile.bounds.intersects(currentBounds)){
                    _tiles_displayed.addLayer(tile);
                }else{
                    // else -> remove from its .cityLayer
                    _tiles_displayed.removeLayer(tile);
                }
            }

        }

    });

    L.SeMap.epsg3857ToLatLng = function(xy){
        var _map = this;
        if(Array.isArray(xy) && Array.isArray(xy[0])){
            return xy.map(function(xy1, idx1){
                return L.SeMap.epsg3857ToLatLng(xy1);
            })
        }else{
            return L.CRS.EPSG3857.unproject(L.point(xy));
        }
    };

    L.SeMap.switchCoordinates = function(xy){
        var _map = this;
        if(Array.isArray(xy) && Array.isArray(xy[0])) {
            return xy.map(function (xy1) {
                return L.SeMap.switchCoordinates(xy1);
            })
        }else{
            return [xy[1], xy[0]];
        }
    };

    L.se_map = function(id, options){ return new L.SeMap(id, options); };




}(window));
